# For our illustration of coroutines for co-operative multitasking we
# will construct a package, `task`, that implements
# the actor model. Our concurrent units of computation are called
# tasks. They are implemented as coroutines and voluntarily give up
# control at appropriate times by yielding to a dispatcher which
# then gives other tasks a chance to run. Tasks may
# be completely independent of each other or may interact through
# _messages_.
#
# Our package is rudimentary and is meant to highlight key considerations
# involved in implementing a co-operative multitasking framework in Tcl.
# Amongst other things, a production level package will need to be
# both more featureful and more robust with better error handling and recovery,
# resource limits on message queues, a more sophisticated scheduling 
# algorithm to prevent starvations and so on.

# We start off by importing the `coroutine` package from
# {tclliblink} and defining our package namespace.
package require coroutine
namespace eval task {
    variable tasks {}
}

# [discrete]
# ===== Creating a new task
# The `task::task` command creates a new task in which the provided
# script will be evaluated.

proc task::task {script} {
    variable next_id
    variable tasks
    variable tasks_added
    
    set task_id [namespace current]::task#[incr next_id]
    dict set tasks $task_id State INIT
    dict set tasks $task_id Messages {}
    set tasks_added 1
    coroutine $task_id apply {script {
        yield
        try $script finally { task::cleanup [task::myself] }
    }} $script
    wakeup_dispatcher
    return $task_id
}
# The `tasks` variable is a dictionary of active tasks
# keyed by task identifiers which are dynamically generated using the
# `next_id` counter.  The task identifiers also serve as the name of
# the coroutine hosting the task. 
# To create a new task, the command initializes the state of the task
# to `INIT` with an empty `Messages` that will hold any messages sent to it.
# It then creates a coroutine context in which to run the supplied script.
# As a general principle, we prefer to run the script within an anonymous
# procedure so as to avoid potentially polluting the global namespace.
#
# The coroutine immediately yields on entry rather than executing the
# task script. The reason for this is that we want all execution to be
# controlled by the dispatcher. Moreover, we do not want the current task
# to be preempted by the new one unless it voluntarily gives up control
# explicitly.
#
# Just before returning, we wake up the task dispatcher in case it
# was previously suspended.

# [discrete]
# ===== Task identity
# A script often needs to know the task under which it is running.
# The `myself` command returns the id of the current task.
# If called from outside a task, it raises an exception.
proc task::myself   {} {
    variable tasks
    set me [info coroutine]
    if {[dict exists $tasks $me]} {
        return $me
    }
    error "Not running in a task."
}

# [discrete]
# ===== Task cleanup
# Any structures allocated by our package need to be cleaned up
# when the task exits.
# The `cleanup` command is more or less self explanatory.
proc task::cleanup {task_id} {
    variable tasks
    if {[dict exists $tasks $task_id]} {
        catch {rename $task_id ""}
        dict unset tasks $task_id
    }
}
# Note that this does not release any resources allocated by the
# task itself, for example open channels. That is the responsibility
# of the task script.

# [discrete]
# ===== Sending messages
# Now we start coming to the meat of the package, implementation of messages
# and dispatching of tasks. Our package itself does not care about the
# content of a message. Its interpretation is left entirely
# to the receiving task. Sending a message involves placing it on the
# task's message queue and waking up our task dispatcher in case it
# is not running. By design, we choose to ignore messages sent to
# non-existing tasks.
proc task::send {task_id message} {
    variable tasks
    if {[dict exists $tasks $task_id]} {
        dict with tasks $task_id {
            lappend Messages $message
        }
        wakeup_dispatcher
    }
    return
}

proc task::wakeup_dispatcher {} {
    variable dispatcher_alarm
    set dispatcher_alarm 1
}
# We will explain `wakeup_dispatcher` later when we describe the dispatcher
# implementation.

# Note that as stated earlier message passing is asynchronous in our framework.
# The sender does not wait for the receiver to accept or process the message.

# [discrete]
# ===== The message queue
#
# It is worth digressing a little bit at this point to elaborate
# on a design choice we made to
# use an explicit message queue as opposed to alternative implementations
# that you might find being used elsewhere.
#
# The first of these alternatives dispenses with queueing of messages altogether.
# Sending a message to a task is implemented by calling the
# corresponding coroutine, either directly or via `yieldto`, passing
# the message as an argument. You might think of this as a _push_
# model where messages are pushed to the task whereas ours is a _pull_
# model where messages are explicitly read by the task with the `receive`
# command described later. There are two problems
# with the push method in a general purpose package like ours:
#
# * The task (coroutine) might have yielded for reasons other than
# for the purpose of receiving messages. For example, the task may use
# calls such as `coroutine::util gets` or `coroutine::util after` that
# we saw in previous sections. In such cases, it expects to be resumed
# from the `gets` I/O handler or from the `after` timer implementation.
# Being resumed with a message argument at that point instead of the
# result of a `gets` would completely confuse the poor coroutine.
#
# * Directly pushing messages to coroutines can lead to issues of
# reentrancy that can be difficult to diagnose and resolve.
#
# The other implementation you might see is the use of Tcl's event queue
# for passing messages instead of the explicit queue in our design. This
# also is essentially a _push_ model where it is the event dispatcher that is
# pushing messages to the tasks. This does not have the reentrancy problems
# but suffers from the first issue noted above.
#
# Maintaining an explicit message queue where tasks pull messages avoids
# these issues. It also has the added benefit of allowing more sophisticated
# capabilities such as message priorities, 
# pattern matching etc.

# [discrete]
# ===== Receiving messages
# The complementary command to `send` is `recv` which returns the next
# message from the tasks message queue.
proc task::recv {} {
    variable tasks
    set me [myself]
    while {1} {
        set msgs [dict get $tasks $me Messages]
        if {[llength $msgs] != 0} {
            dict set tasks $me Messages [lassign $msgs msg]
            return $msg
        } else {
            dict set tasks $me State RECEIVE
            suspend
        }
    }
}
# The implementation is again more or less obvious. If the message queue
# is not empty, the first message is removed and passed back to the task.
# Otherwise, the task suspends itself thereby allowing other tasks to
# run.
#
# Note that in our simple implementation, if a task continuously receives
# messages, it will not allow other tasks to run. In production quality
# software, a limit on the number of messages processed before suspension
# would be well advised. It could be up to the task to do this or our
# package could be enhanced to keep track of the number of messages
# received since the last suspension and suspend the thread even if
# the message queue is not empty.

# [discrete]
# ===== Relinquishing the CPU
# There may be tasks that run independently without sending or receiving
# messages. Being good citizens however, they may wish to relinquish
# the CPU to other tasks. The `relinquish` command lets them do that.
proc task::relinquish {} {
    variable tasks
    dict set tasks [myself] State RELINQUISHED
    suspend
}
# We will see the purpose of the state `RELINQUISHED` later.

# Suspending a task is nothing more than a simple `yield`.
proc task::suspend {} {
    yield
}

# [discrete]
# ===== The task dispatcher
# We finally come to the heart of our package, the dispatcher. The role
# of the dispatcher is to keep looping through the list of tasks looking
# for any that have messages pending in their receive queue and invoking
# them. There are several finer points that need discussion but
# first we show the code.

# Our dispatcher is itself a coroutine whose innards are
# implemented by the `dispatch_loop` procedure.
# This will loop continuously initializing some state
# variables and then running an inner loop that iterates over the
# *current* tasks invoking those that are ready to run. There is a
# subtle point here in that tasks may be added or removed by any of
# the current tasks. We will not schedule any new tasks to run until
# the current tasks have been invoked first.  This is a matter of
# policy; you may choose otherwise.
#
# A task is considered
# ready to run if
#
# * it is a new task indicated by the `INIT` state.
# * it has voluntarily given up the CPU and is awaiting its next turn
# as indicated by the `RELINQUISHED` state.
# * it is waiting for messages as indicated by its `RECEIVE` state *and*
# there are messages pending in its queue.
#
proc task::dispatch_loop {} {
    variable tasks
    variable dispatcher_alarm
    variable tasks_added
    
    while {1} {
        set woken 0
        set tasks_added 0
        foreach task_id [dict keys $tasks] {
            if {! [dict exists $tasks $task_id]} continue
            dict with tasks $task_id {}; # Copy dictionary keys to local variables
            if {($State in {INIT RELINQUISHED}) ||
                ($State eq "RECEIVE" && [llength $Messages])} {
                incr woken
                dict set tasks $task_id State RUNNING
                if {[catch { $task_id } mesg]} {
                    puts stderr "$task_id ended with error $mesg"
                    cleanup $task_id
                }
            }
        }
        if {$woken == 0 && $tasks_added == 0} {
            coroutine::util vwait [namespace current]::dispatcher_alarm
        }
    }
}
# We need to make an important point about the task's state -- why
# we mark a task as `INIT`, `RELINQUISHED`, `RECEIVE` or `RUNNING`.
# This gets back to our earlier discussion of the push versus pull model.
# A task in the `RUNNING` state may actually have yielded outside of our
# framework through a call such as `coroutine::util gets`. We must not
# invoke such tasks from our dispatcher for the same reasons discussed
# there. Maintaining state information allows us to eliminate these
# `RUNNING` tasks from consideration. The `RECEIVE` and 
# `RELINQUISHED` states are distinguished by the former indicating
# the task should be awoken *only* if it has messages waiting, while
# the latter has no such requirement.
#
# Invocation of a task is done by transferring control to its hosting
# coroutine. In case of any exceptions, we delete
# the task from the task database.
#
# [NOTE]
# Another option for transferring control to the task coroutine would
# have been to use `yieldto` instead of the direct call. 
# You might find this used in some implementations. We prefer to
# call the task coroutine directly because our dispatcher would then
# regain control when the coroutine does a `yield`. This is important
# for some edge cases where the task yields but not using one of our
# package commands. For example, a task that sits in a loop using
# `coroutine::util gets` or the like would relinquish control but
# not to the dispatcher.
#
# The other detail we need to take care of is with regards to the
# actions to take when none of the tasks are in a ready condition.
# Clearly we do not want the dispatcher to just spin endlessly without
# permitting other code to run. We would like it to suspend itself
# until there is a reason for it to run. We have a couple of options
# here. We can pass control to Tcl's event loop after scheduling
# ourselves to run after some short period of time. Instead of this
# polling mechanism, we choose the other option: we wait via the
# `coroutine::util vwait` call for a variable, `dispatcher_alarm` to
# be set. Correspondingly, the `wakeup_dispatcher` procedure, called
# on a `send` to any task, sets this variable. If the dispatcher is already
# active, it has no effect; otherwise, it releases the dispatcher from
# its suspended state.
#
# Next we have to consider when the dispatcher should suspend itself via
# the `vwait`. Two conditions should be met for this:
#
# * No tasks should be ready to run. We track this with the `woken`
# variable which is reset at the top of every dispatcher iteration and
# incremented every time a task is run. If this is `0` at the end of
# a complete loop through the task list, we know no tasks are ready.
# * No new tasks must have been created since we initiated the iteration
# through the tasks. Again we use a variable, `tasks_added`, to keep track
# of this. The variable is reset at the top of the iteration and set
# when a new task is created within the `task` procedure.
#
# The dispatcher suspends itself when both the above conditions are
# satisfied.

# [discrete]
# ===== Starting the dispatcher
# We have defined our dispatcher loop procedure but need to provide a
# means for the application to create it. So we do that now.
proc task::start_dispatcher {} {
    coroutine dispatcher dispatch_loop
}

# Our package is now ready for use.
package provide task 1.0
