# For demonstration purposes, we will implement a little internal
# DSL that an accountant might use to generate paychecks. The
# following program is a sample that an accountant expects to be run
# against every employee record in a database to generate paychecks.
#
# ----
# deduct insurance 100
# deduct federal 10% when salary between 20000 and 30000
# deduct federal 20% when salary above 30000
# deduct state 5% when federal above 2500 <1>
# generate paycheck
# generate paystub
# ----
# <1> Yes, silly I know but that's taxes for you
#
# Our little language should be understandable to an accountant right
# away. Most hosting languages would need to implement the above as
# external DSL's. Not so in Tcl because *it is standard Tcl syntax*.
#
# We start with a couple of utility procedures to check whether a
# word is included in a permitted list and to ensure a word is an integer.
# Nothing to see here.
proc check_word {word allowed} {
    if {$word ni $allowed} {
        error "Invalid word '$word'. Must be one of [join $allowed ,]"
    }
}
proc check_number {word} {
    if {![string is integer -strict $word]} {
        error "'$word' is not an integer"
    }
}

# We will assume employee data is passed to us in a dictionary.
# We will need a procedure to map the keywords in our DSL to keys
# in the employee dictionary.
proc word_to_var {word} {
    dict get {
        salary  Salary
        federal FederalTax
        state   StateTax
        insurance Insurance
    } $word
}

# As we come to the meat of the implementation, we have a decision to make.
# The obvious design is to map the commands in the DSL to commands in the
# safe interpreter that are aliased in the master. At that point, we need to
# make choice -- should the DSL be interpreted or compiled to a Tcl script.
# In the first case, our alias implementations would be called for every
# employee record. They would then take the appropriate actions to 
# deduct taxes, print the check and so on. We will go with the second 
# alternative and compile the DSL to a Tcl script. This Tcl script will
# then be directly evaluated for each employee record. The advantage
# is that the DSL program need not be repeatedly parsed for every employee
# record. This method is slightly more complex that the interpreted DSL
# implementation but since you have already seen metaprogramming in
# <<sect_code_meta>>, it should not be hard to follow.

# We create a class to store some context -- the compiled Tcl script
# for the DSL code.
oo::class create Paymaster {
    variable Script
}

# Our paymaster needs to be told the rules for generating a paycheck.
# These rules are provided by our accountant using our DSL. We will
# pass these in to our paymaster through its constructor. 
oo::define Paymaster constructor {dslscript} {
    interp create dslengine -safe
    dslengine eval {namespace delete ::}; # Deleting the global namespace will remove all commands from the slave interpreter.
    foreach alias {deduct generate} {
        dslengine alias $alias [self] $alias
    }
    dslengine eval $dslscript
    interp create calculator -safe
    calculator alias puts puts
}

# Our constructor needs some explanation. We create not one, but two,
# slave interpreters. The first one, `dslengine`, is fundamental to our
# design because it is the one that will parse the DSL script. We remove
# *all* commands from it and then add the aliases `deduct` and
# `generate` corresponding to the DSL language commands. These then will
# be the only commands accessible in that slave.
#
# The second interpreter is not strictly necessary. Once `dslengine`
# has generated the Tcl script, we could run that in the master interpreter.
# Its presence is to add another layer of security. We do not trust our
# accountant. He may be annoyed about his annual raise and create a rule
# that "`deducts`" `[exec reformat drive]` instead of a number. Although
# our aliases will validate arguments, there is a chance that something
# will slip through. Running the generated script in another safe
# interpreter protects against oversights. Note we cannot use our
# `dslengine` for this because it does not have any commands other than
# the DSL-specific ones.

# Next we define our destructor which has nothing to do but release the slaves.
oo::define Paymaster destructor {
    catch { interp delete dslengine }
    catch { interp delete calculator }
}

# Time now to define our aliases that implement the DSL. Only two are
# needed for our simple language. We will start with the more complex one,
# `deduct`. Remember our design involves *transpiling* the DSL into Tcl
# so these methods need to generate fragments of Tcl that reflect
# the DSL statements. These fragments are concatenated to construct the
# transpiled Tcl script.
#
# We have already seen techniques for constructing scripts in
# <<sect_code_constructing_scripts>>. We follow that method here by
# defining a script template that reflects the logic of the
# script fragment implementing the `deduct` DSL statement. We then
# construct the script fragments for each place holder in the template
# and then replace them at the end with `string map`. We will not go
# into further detail here as that is not the point of this exercise.
oo::define Paymaster method deduct {what deduction args} {
    check_word $what {federal state insurance}
    set category [word_to_var $what]
    if {[regexp {^(\d+)(%)?$} $deduction -> amount percent] == 0} {
        error "Invalid deduction amount $deduction"
    }
    set template {
        if {%CONDITION%} {
            if {%USE_PERCENT%} {
                set deduction [expr {int(($Salary * %AMOUNT%)/100)}]; # Deduction is a percentage
            } else {
                set deduction %AMOUNT%; # Deduction is an absolute value
            }
            incr %CATEGORY% $deduction; # Allocate deduction to a category...
            incr NetAmount -$deduction; # ...and remove net amount accordingly
        }
    }

    if {[llength $args] == 0} {
        set condition true; # No condition for the deduction
    } else {
        if {[llength $args] < 4} {
            error "Incomplete line"
        }
        set args [lassign $args cond_keyword cond_var cond_cmp number]
        check_word $cond_keyword {when if}
        set cond_var [word_to_var $cond_var]
        check_number $number
        switch -exact -- $cond_cmp {
            over - above {
                set condition "\[set $cond_var\] > $number"
            }
            under - below {
                set condition "\[set $cond_var\] < $number"
            }
            within - between {
                if {[llength $args] != 2 ||
                        [lindex $args 0] ne "and"} { 
                    error "Invalid \"$cond_cmp\" arguments"
                }
                set upper [lindex $args 1]
                set condition "\[set $cond_var\] > $number && \[set $cond_var\] < $upper"
            }
        }
    }        

    append Script [string map [list \
                                   %CATEGORY% $category \
                                   %CONDITION% $condition \
                                   %AMOUNT% $amount \
                                   %USE_PERCENT% [expr {$percent ne ""}]] \
                       $template]
}

# The other DSL statement is `generate`. The generation of the script
# fragment here is much simpler.
oo::define Paymaster method generate {what} {
    switch -exact -- $what {
        paycheck {
            append Script {puts "Pay to $Name the amount of $NetAmount"} \n
        }
        paystub {
            append Script {
                puts "Paystub: $Name Salary=$Salary Net=$NetAmount"
                puts "         Fed=$FederalTax State=$StateTax Insurance=$Insurance"
            } \n
        }
        default { error "No means of generating \"$what\"" }
    }
}

# Now we define the method that will be called by the application to pay an
# employee. The employee record is passed in the form of a dictionary
# with the employee's name and salary. It then executes the generated
# Tcl script within the scope of the dictionary in our `calculator` slave.
oo::define Paymaster {
    method pay {emp} {
        set emp [dict merge {
            FederalTax 0
            StateTax 0
            Insurance 0
        } $emp]
        dict set emp NetAmount [dict get $emp Salary]
        calculator eval [list set emp $emp]
        calculator eval [list dict with emp $Script]
    }
}

# Our final method is just for the purposes of inspecting the Tcl
# script we generated. Just so we can see what havoc we have wrought.
oo::define Paymaster method script {} { return $Script }
