# Our `memfs` package will implement an in-memory virtual file
# system. Mounting a `memfs` will create a new VFS instance with no
# content. Any content written to this file system will be erased
# when the VFS instance is unmounted. An application may mount multiple
# `memfs` VFS instances, each independent of the others.

# We require the `vfs` package which exposes Tcl's VFS features at the
# script level. We will also need the `tcl::chan::variable` package
# which we will use to implement channels targeting our VFS.

package require vfs
package require tcl::chan::variable

#adocgen off

#
# We will store our in-memory file system as a nested dictionary.
#
# A _directory node_ will have nested keys
# `Files`, `Directories`, `Attributes` and `Stat`. The `Files`
# key will contain a nested dictionary containing names of plain files
# within that directory each of which will hold a nested dictionary node
# corresponding to a file. The `Directories` key will contain a nested
# dictionary containing names of subdirectories each holding a nested
# dictionary node corresponding to a directory. The 
# `Stat` contains information about access time etc.
#
# A _file node_ is similar except that instead of the `Files` and
# `Directories` keys, it contains a `ContentVar` key which holds the
# name of a variable containing the file content.
# 

#adocgen on

# We will choose to implement our VFS using namespaces as we did for
# our virtual channel example and define an ensemble
# corresponding to the VFS driver subcommands.
namespace eval memfs {
    namespace ensemble create -parameters {fs_id} -subcommands {
        access createdirectory deletefile fileattributes
        matchindirectory open removedirectory stat utime
    }
}

# [NOTE]
# A minor point is worth noting about our namespace ensemble
# definition.  Since we support multiple instances of our VFS, we need
# to pass an instance identifier to these subcommand procedures in
# addition to the arguments passed by the VFS core. The VFS core takes
# a single command prefix and appends its own arguments to it. Our
# instance identifier will be part of the command prefix and hence
# will appear *before* the subcommand.  We define the ensemble
# accordingly with the `-parameters` option to indicate the subcommand
# position. See <<ns#sect_ns_ensembles>> for
# details.

# === Signalling VFS errors
# Let us first deal with signalling of errors as the VFS subsystem
# expects implementations to use a specific call for this purpose
# as opposed to directly raising exceptions using the standard
# Tcl `error` or `throw` commands. This is to ensure
# a consistent set of error messages and error codes irrespective
# of the underlying file system.
#
# Errors within the VFS driver should be signalled by calling
# the `vfs::filesystem posixerror` command and passing it a numeric
# code representing a POSIX error. This command will then raise
# a standard Tcl exception with an appropriately formatted error code.
# Since numeric codes are hard to remember, we will define a wrapper,
# `posix_error`, that will also accept the mnemonic equivalent of
# numeric error codes.
proc memfs::posix_error {err} {
    if {![string is integer -strict $err]} {
        set err [::vfs::posixError $err]
    }
    vfs::filesystem posixerror $err
}
# So for example, we can call
# ----
# posix_error ENOENT
# ----
# to report an error that a file or directory does not exist.

# === Mounting and unmounting
#
# Following `vfs` package conventions, we will name our command
# for mounting a `memfs` VFS as `Mount`. 
proc memfs::Mount {mount_path} {
    set id [init_fs]
    vfs::filesystem mount $mount_path [list [namespace current] $id]
    vfs::RegisterMount $mount_path [list [namespace current]::Unmount $id]
    return $id
}
# It takes a single argument which is the mount point where our file
# system will be located. It calls an internal command `init_fs` to
# create a new file system. The `vfs::filesystem mount` command then
# mounts the file system at the specified location. It has the general
# syntax
# [subs="normal"]
# [source,Tcl]
# ----
# vfs::filesystem mount ?-volume? _PATH_ _CMDPREFIX_
# ----
# Here `_PATH_` is the mount point and _CMDPREFIX_ is the command prefix
# that implements the VFS driver. In our example, this is the ensemble
# command we defined earlier which has the same name as our implementation
# namespace. Notice we also pass in the `memfs` instance identifier
# returned by `init_fs` command. This allows our implementation to
# distinguish between multiple `memfs` instances.
#
# The `-volume` prefix specifies that the file system is also
# a new volume as seen by the `file volumes` command. We saw an example
# of such a VFS with the `vfs::urltype::ftp` example earlier. This option
# must not be specified for file systems, like `memfs` that will mount
# within an existing native file system path.
#
# The call to `vfs::RegisterMount` is strictly not necessary. However,
# it allows the application to unmount out VFS by passing the mount point to the
# generic `vfs::unmount` command instead of having to call our VFS-specific
# `Unmount` command.

# Next we implement the corresponding command for unmounting a `memfs`
# VFS instance.
proc memfs::Unmount {fs_id mount_path} {
    variable file_systems
    if {![info exists file_systems($fs_id)]} {
        return
    }
    vfs::filesystem unmount $mount_path
    unset file_systems($fs_id)
    namespace delete $fs_id
    return
}
# The main point to note here is the call to `vfs::filesystem unmount`.
# The rest of the code is specific to our implementation and essentially
# deletes all data associated with that VFS instance.
# Instead of directly calling our `Unmount` command, it is recommended
# applications should call `vfs::unmount` so that it can update its
# database of mounted file systems.
#
# [WARNING]
# In any case, applications should not call the `vfs::filesystem unmount`
# command themselves. That command will remove the file system from Tcl's
# view but will *not* notify the VFS driver that the file system has been
# unmounted.

# === VFS operations
#
# We now move on to implementation of the driver commands called by
# the Tcl VFS subsystem for operating on a file within the VFS. The
# first four arguments to all these commands are identical. The first
# is the `memfs` VFS instance id (which we passed as part of the
# command prefix), the mount point path, the relative path within the
# VFS, and the original path as specified in the command that invoked
# the file operation. For example, if our VFS was mounted at
# `/tmp/mem` and the current working directory was `/tmp`, then
# a command like
# [source, Tcl]
# ----
# file exists mem/foo.txt
# ----
# would result in the arguments after the `memfs` instance id being
# `/tmp/mem`, `foo.txt` and `mem/foo.txt` respectively.

# ==== Checking access: `access`
#
# We start off with implementation of the `access` command. This is
# used by the VFS subsystem to check if a specific file or directory
# can be accessed with the specified mode. The mode is passed
# as an additional argument in the form of
# a bitmask that indicates the type of desired access. If `0`,
# only existence of the file is to be checked. The low three bits
# signify execute (least significan bit), write and read access respectively.
# Our file system does not implement access permissions and
# so for directories we will allow all access types. For files however,
# we will disallow execute access since the operating system has no
# knowledge of our file system and cannot execute files residing in it.
# (At the Tcl level, `exec` will not work.)
proc memfs::access {fs_id root relpath origpath mode} {
    switch -exact -- [node_type $fs_id $relpath] {
        "" { posix_error ENOENT }
        file { if {$mode & 1} { posix_error EACCES } }
        dir { }
    }
    return
}
# The implementation uses our internal `node_type` command to check for
# whether the path is a regular file or directory. If the specified access is
# allowed the command returns normally with the return value being
# immaterial. If access is *not* allowed, the command must raise
# a POSIX error as discussed previously. In our case, we return `ENOENT`
# when the path does not exist and `EACCES` when it does not have
# the requested execute access permission.

# ==== Creating directories: `createdirectory`
# Next we implement the `createdirectory` call which is straightforward.
# The command is expected to create a new directory at the specified path
# within our file system if it does already exist. If the path corresponds
# to an existing regular file, it should raise a POSIX error.
proc memfs::createdirectory {fs_id root relpath origpath} {
    node_add_dir $fs_id [node_find $fs_id $relpath dir]
}
# The implementation uses two internal commands. The first of these,
# `node_find`, maps a file path to the location key for the corresponding node
# in our internal file system structures. The node itself need not
# exist but if it does, the optional third argument mandates that it
# must be of the specified type. Thus in the above call, `node_find`
# would raise a POSIX error if the node existed and was not a directory.
# We will see `node_find` used throughout our implementation.
#
# The other internal command, `node_add_dir` simply creates an empty
# directory and associated structures at a specified node location.

# ==== Deleting directories: `removedirectory`
# The `removedirectory` command deletes a directory. It takes an
# additional argument which must be a boolean value. If `true`
# then the directory and its contents should be recursively deleted;
# otherwise the command should raise a POSIX error if the directory
# is not empty. It is not an error if the directory does not exist.
proc memfs::removedirectory {fs_id root relpath origpath recursive} {
    node_del_dir $fs_id [node_find $fs_id $relpath dir] $recursive
}
# The implementation is pretty much identical to that of
# `createdirectory` above except we call `node_del_dir` which does
# the hard work of deleting the directory structure and its contents.

# ==== Creating and opening files: `open`
# The `open` VFS driver subcommand essentially has to implement the
# file system level operations of the Tcl `open` and `chan open`
# commands which applications use to create files as well as open
# them for I/O. The command takes two additional arguments
# that specify the access mode and permissions in the case of
# creating a new file. These are equivalent to the ones passed
# to the Tcl `open` command.
proc memfs::open {fs_id root relpath origpath mode perms} {
    variable file_systems
    
    set node_key [node_find $fs_id $relpath file]
    set exists [expr {[node_type $fs_id $relpath] ne ""}]
    set truncate 0
    switch -glob -- $mode {
        "" -
        r* {
            if {! $exists} {
                posix_error ENOENT; # File must exist
            }
        }
        a* -
        w* {
            if {$exists} {
                if {[string index $mode 0] eq "w"} {
                    set truncate 1
                }
            } else {
                node_add_file $fs_id $node_key
            }
        }
        default {
            error "Unsupported mode \"$mode\""
        }
    }
    set chan [node_add_channel $fs_id $node_key $truncate]
    set close_callback [list [namespace current]::node_close_handler  \
                            $fs_id $node_key $chan]
    return [list $chan $close_callback]
}
# The implementation of `open` is a little longer than others but
# should be straightforward to follow at this stage. We have already
# described the `node_find` and `node_type` internal commands.
# The `mode` parameter, as for the Tcl `open`, specifies the open mode
# in the form of `r`, `r+`, `w` etc. and the `switch` statement
# takes appropriate action depending on the mode.
# The `node_add_file` command creates a new node of type file in our
# file system structure.
#
# The return value of the command should be a list of one or two elements.
# The first element must be the handle to an open channel to use for
# performing I/O on the file. The second element is optional. If present
# it should be command prefix to be invoked when the channel is closed.
# Our internal `node_add_channel` command creates a new channel to a
# specified file (node). It also adds the channel to the list of channels
# internally associated with the file since we want to prevent
# files from being deleted while they have channels open to them.
# For the same reason, we also need to know when a channel is closed
# so that we can remove it from this list. Thus we make use of
# the second optional element in the return value to declare a callback
# to be invoked on channel close. This callback `node_close_handler`
# will remove the channel from the channel list and also update the
# access and modification time stored for the node (file).

# ==== Deleting files: `deletefile`
# The `deletefile` command is almost identical to the `removedirectory`
# command we implemented above except it deals with
# removal of files, not directories, and hence has no need for
# recursion control.
proc memfs::deletefile {fs_id root relpath origpath} {
    node_del_file $fs_id [node_find $fs_id $relpath file]
}
# A design decision we have made is that files with
# open channels to them cannot be deleted. This follows the Windows
# model (as opposed to Unix). Our `node_del_file` implementation
# will report an error via `posix_error` on an attempt to
# delete a file which is open.
# Naturally, this decision affects `removedirectory` as well.

# ==== Setting file timestamps: `utime`
# The `utime` command is called to set the last access and
# modification timestamps for a file or directory. It takes two
# additional arguments corresponding to the access and modification
# time respectively. These are specified in terms of the number of
# seconds since the epoch, January 1, 1970.
proc memfs::utime {fs_id root relpath origpath atime mtime} {
    node_set_times $fs_id [node_find $fs_id $relpath] $atime $mtime
}
# The internal commands in our implementation set these timestamps
# for various operations. For example, creation of a file will also
# update the modification time for its parent directory. The `utime`
# command is specifically called by Tcl in response to the
# `file atime` and `file mtime` commands.

# ==== File statistics: `stat`
# The `stat` command is returns information about a file or directory.
# The return value from the command should be a dictionary with
# the following keys: `dev`, `ino`, `mode`, `nlink`, `uid`, `gid`,
# `size`, `atime`, `mtime`, `ctime` and `type`. These keys have
# exactly the same semantics we described for the `file stat` command.
proc memfs::stat {fs_id root relpath origpath} {
    return [node_stat $fs_id [node_find $fs_id $relpath]]
}
# Our node implementation maintains the relevant information internally
# and our driver API can again just delegate to it.

# ==== File attributes: `fileattributes`
# As discussed in the <<files#chap_files, Files and Basic I/O>>
# chapter, files can be associated with certain attributes that are
# specific to the file system. These attributes are managed with
# the <<files#sect_files_file_attributes, `file attributes`>> commmand.
# VFS drivers need to implement a corresponding `fileattributes`
# command to allow Tcl to access attributes for files within the VFS.
# The command is invoked in three forms:
#
# * If no additional arguments (other than the standard ones) are
# specified, the command should return a list of attribute names
# supported by the file system. *Every such call must return the
# same names in the same order.*
#
# * If a single additional argument is specified, it will be an
# integer index into the list of attribute names. The command should
# return the value of this attribute.
#
# * If two additional arguments are specified, the first is the integer
# index into the attribute list as above. The second is the value
# to assign to the attribute.
#
# Our implementation is shown below.
proc memfs::fileattributes {fs_id root relpath origpath args} {
    set attr_names [lsort [node_attr_names]]
    if {[llength $args] == 0} {
        return $attr_names
    }
    set node_key [node_find $fs_id $relpath]
    set attr_name [lindex $attr_names [lindex $args 0]]
    if {[llength $args] == 1} {
        return [node_attr $fs_id [node_find $fs_id $relpath] $attr_name]
    } else {
        return [node_attr $fs_id [node_find $fs_id $relpath] $attr_name [lindex $args 1]]
    }
}
# As always, it relies on the underlying internal functions to to the
# actual work. The `node_attr_names` command returns a list of
# attribute names supported by our VFS. We ensure we always pass them
# back in the same order by sorting them before returning. We then use
# the `node_attr` command to get or set the attribute on the node
# corresponding to the specified path.
#
# Although not shown above, our VFS supports two attributes:
# `-contenttype` and `-encoding`. Applications can assign any values
# they want to these attributes as VFS has no interest or control over
# their semantics. The *intended* use is for applications to store the
# encoding used for the file content in the `-encoding` attribute and
# the content type (similar to the `Content-Type` HTTP header) which
# specifies the format of the content like `text/html` in the
# `-contenttype` attribute.  Note these are advisory attributes and
# have to be explicitly set by the application There is no way for the
# VFS to detect the encoding in use or the type of content stored in
# the file.

# ==== Matching files: `matchindirectory`

# One final VFS driver command remains to be implemented. The
# `matchindirectory` command is what Tcl's file system calls to
# retrieve directory contents. The return value of the command is
# expected to be a (possibly empty) list containing file names.
# Commands like `glob` also support matching based on patterns and
# file types.  Correspondingly, the `matchindirectory` command takes
# two additional arguments that specify a glob pattern and a file type
# specifier.
#
# If the glob pattern is the empty string, the `relpath`
# argument is the relative path of a file or directory. The command
# should then only check if the path exists *and* is of the specified
# type. If so, it should return a list containing the corresponding
# *original path* that was specified on the command line, not the
# relative path. If the path does not exist or is not of the specified
# type, an empty list is returned.
#
# If the glob pattern is not the empty string, the `relpath` is
# always a path to an existing directory whose contents are to be
# matched against the pattern. The command should then return the names
# that match the pattern and the specified type.
proc memfs::matchindirectory {fs_id root relpath origpath pat type} {
    variable file_systems
    if {[string length $pat] == 0} {
        set file_type [node_type $fs_id $relpath]
        if {($file_type eq "dir" && [::vfs::matchDirectories $type]) ||
            ($file_type eq "file" && [::vfs::matchFiles $type])} {
            return [list $origpath]
        } else {
            return {}
        }
    }

    if {[node_type $fs_id $relpath] ne "dir"} {
        return {}
    }
    
    set node_key [node_find $fs_id $relpath]
    set matches {}
    if {[::vfs::matchDirectories $type]} {
        foreach name [node_subdirs $fs_id $node_key] {
            if {[string match $pat $name]} {
                lappend matches [file join $origpath $name]
            }
        }
    }
    if {[::vfs::matchFiles $type]} {
        foreach name [node_files $fs_id $node_key] {
            if {[string match $pat $name]} {
                lappend matches [file join $origpath $name]
            }
        }
    }
    return $matches
}
# Our implementations makes use of two internal commands `node_files`
# and `node_subdirs` that return the names of files and subdirectories
# under the specified node. These are then matched against the pattern
# to construct the returned list.
#
# The `type` argument specifies whether the returned list should
# include files, directories or both. We treat it as opaque and use
# the utility commands `vfs::matchDirectories` and `vfs::matchFiles`
# to ascertain whether entries of a specific type are to be included
# or not. Note that the type argument can indicate that *both* files
# and directories are to be included.
#
# [NOTE]
# The Tcl `glob` command allows the type argument to limit matching
# files based on other criteria as well such as permissions. These criteria
# are not exposed for VFS systems.

# The `vfs` package provides another utility command `vfs::matchCorrectTypes`
# that serves as an alternative to `matchDirectories` and `matchFiles`.
# [subs="normal"]
# [source,Tcl]
# ----
# vfs::matchCorrectTypes _TYPES_ _FILELIST_ ?_DIR_?
# ----
# The command returns a list of names from `_FILELIST_` that
# fulfil the type requirements specified by `_TYPES_`. If `_DIR_` is
# not specified, `_FILELIST_` must contain absolute paths. If `_DIR_`
# is specified, it must be a directory and `_FILELIST_` should be
# the list of file and directory names within that directory. Depending
# on your internal file system implementation, you may find
# `matchCorrectTypes` more convenient to use.




# ==== `memfs` internals
# Due to space limitations, we will not go into detail regarding our
# `node_*` commands that implement our VFS internals. You can download
# the `memfs.tcl` file from the book's web site to see the implementation.

#adocgen off

proc memfs::init_fs {} {
    variable file_systems
    variable fs_id_counter
    
    # Note we are using the filesystem id as the device number
    # for the stat call. It should therefore be an integer value.
    set fs_id [incr fs_id_counter]
    
    set file_systems($fs_id) [dict create Inode 0]
    set node [node_create $fs_id dir]
    namespace eval $fs_id {};   # Used to hold content variables
    dict set file_systems($fs_id) {} $node
    return $fs_id
}

proc memfs::node_create {fs_id type} {
    variable file_systems

    dict incr file_systems($fs_id) Inode
    set inode [dict get $file_systems($fs_id) Inode]
    if {$type eq "file"} {
        set node [dict create Attrs {} Chans {} \
                      ContentVar [namespace current]::${fs_id}::content$inode]
    } else {
        set node [dict create Directories {} Files {} Attrs {}]
    }

    set now [clock seconds]
    set stat [list \
                  atime $now \
                  ctime $now \
                  mtime $now \
                  mode 0777  \
                  dev $fs_id \
                  ino $inode \
                  nlink 1 \
                  uid -1 \
                  gid -1]
    dict set node Stat $stat
    return $node
}

proc memfs::node_type {fs_id relpath} {
    variable file_systems
    
    set node_key [list {}]
    set parts [file split $relpath]
    if {[llength $parts] == 0} {
        return dir;             # The top level directory
    }
    foreach elem [lrange $parts 0 end-1] {
        lappend node_key Directories $elem
    }

    if {![dict exists $file_systems($fs_id) {*}$node_key]} {
        return ""
    }

    set tail [lindex $parts end]
    if {[dict exists $file_systems($fs_id) {*}$node_key Files $tail]} {
        return file
    }
    if {[dict exists $file_systems($fs_id) {*}$node_key Directories $tail]} {
        return dir
    }

    return ""
}

proc memfs::node_find {fs_id relpath {type ""}} {
    variable file_systems
    
    set node_key [list {}]
    set parts [file split $relpath]
    if {[llength $parts] == 0} {
        return $node_key;        # Root directory
    }
    foreach elem [lrange $parts 0 end-1] {
        lappend node_key Directories $elem
    }

    if {![dict exists $file_systems($fs_id) {*}$node_key]} {
        posix_error ENOENT;     # Parent directories do not exist
    }

    set tail [lindex $parts end]
    dict with file_systems($fs_id) {*}$node_key {
        switch -exact -- $type {
            dir {
                if {[dict exists $Files $tail]} {
                    posix_error EEXIST; # Emulate errors on Windows
                }
                lappend node_key "Directories" $tail
            }
            file {
                if {[dict exists $Directories $tail]} {
                    posix_error EACCES; # Emulate errors on Windows
                }
                lappend node_key "Files" $tail
            }
            "" {
                if {[dict exists $Directories $tail]} {
                    lappend node_key "Directories" $tail
                } elseif {[dict exists $Files $tail]} {
                    lappend node_key "Files" $tail
                } else {
                    posix_error ENOENT;
                }
            }
        }
    }
    return $node_key
}

proc memfs::node_parent {node_key} {
    if {[llength $node_key] < 3} {
        error "Internal error: node_parent called on root node"
    }
    return [lrange $node_key 0 end-2]; # Assumes valid key other than root!
}

proc memfs::node_exists {fs_id node_key} {
    variable file_systems
    return [dict exists $file_systems($fs_id) {*}$node_key]
}

proc memfs::node_attr_names {} {
    return {-contenttype -encoding}
}

proc memfs::node_attr {fs_id node_key attr args} {
    variable file_systems
    if {$attr ni [node_attr_names]} {
        error "Internal error: unknown attribute name $attr"
    }

    dict with file_systems($fs_id) {*}$node_key {
        if {[llength $args] == 0} {
            if {[dict exists $Attrs $attr]} {
                return [dict get $Attrs $attr]
            } else {
                return ""
            }
        } elseif {[llength $args] == 1} {
            dict set Attrs $attr [lindex $args 0]
        } else {
            error "Internal error: invalid # args"
        }
    }
}

proc memfs::node_add_dir {fs_id node_key} {
    variable file_systems
    if {![node_exists $fs_id $node_key]} {
        dict set file_systems($fs_id) {*}$node_key [node_create $fs_id dir]
        node_set_times_now $fs_id [node_parent $node_key]
    }
}

proc memfs::node_add_file {fs_id node_key} {
    variable file_systems
    dict set file_systems($fs_id) {*}$node_key [node_create $fs_id file]
    node_set_times_now $fs_id [node_parent $node_key]
    return
}

proc memfs::node_del_file {fs_id node_key} {
    variable file_systems
    if {![dict exists $file_systems($fs_id) {*}$node_key]} {
        return
    }

    if {[dict size [dict get $file_systems($fs_id) {*}$node_key Chans]] != 0} {
        posix_error EACCES;     # There is at least one channel open to the file.
    }

    unset [dict get $file_systems($fs_id) {*}$node_key ContentVar]
    dict unset file_systems($fs_id) {*}$node_key

    set now [clock seconds]
    node_set_times $fs_id [node_parent $node_key] $now $now
    return
}

proc memfs::node_del_dir {fs_id node_key recursive} {
    variable file_systems

    if {$node_key eq ""} {
        posix_error EACCES;     # Not allowed to delete root of our file system.
    }

    if {![dict exists $file_systems($fs_id) {*}$node_key]} {
        return
    }
    
    if {$recursive} {
        node_del_dir_recurse $fs_id $node_key
    } else {
        if {[dict size [dict get $file_systems($fs_id) {*}$node_key Files]] != 0 ||
            [dict size [dict get $file_systems($fs_id) {*}$node_key Directories]] != 0} {
            posix_error EEXIST
        }
        dict unset file_systems($fs_id) {*}$node_key
        node_set_times_now $fs_id [node_parent $node_key]
    }
    return
}

proc memfs::node_del_dir_recurse {fs_id node_key} {
    variable file_systems

    foreach file [dict keys [dict get $file_systems($fs_id) {*}$node_key Files]] {
        node_del_file [linsert $node_key end Files $file]
    }
    
    foreach dir [dict keys [dict get $file_systems($fs_id) {*}$node_key Directories]] {
        node_del_dir_recurse [linsert $node_key end Directories $dir]
    }
    
    dict unset file_systems($fs_id) {*}$node_key
    node_set_times_now $fs_id [node_parent $node_key]
    return
}

proc memfs::node_add_channel {fs_id node_key truncate} {
    variable file_systems
    dict with file_systems($fs_id) {*}$node_key {
        if {$truncate} {
            set $ContentVar ""
        }
        set chan [::tcl::chan::variable $ContentVar]
        dict set Chans $chan 1
    }
    return $chan
}

proc memfs::node_set_times {fs_id node_key atime mtime} {
    variable file_systems
    dict set file_systems($fs_id) {*}$node_key Stat atime $atime
    dict set file_systems($fs_id) {*}$node_key Stat mtime $mtime
    return
}

proc memfs::node_set_times_now {fs_id node_key} {
    set now [clock seconds]
    node_set_times $fs_id $node_key $now $now
}

proc memfs::node_stat {fs_id node_key} {
    variable file_systems
    dict with file_systems($fs_id) {*}$node_key {
        set stat $Stat
        if {[info exists ContentVar]} {
            dict set stat type file
            dict set stat size [string length [set $ContentVar]]
        } else {
            dict set stat type directory
            dict set stat size 0
        }
    }
    return $stat
}

proc memfs::node_files {fs_id node_key} {
    variable file_systems
    return [dict keys [dict get $file_systems($fs_id) {*}$node_key Files]]
}

proc memfs::node_subdirs {fs_id node_key} {
    variable file_systems
    return [dict keys [dict get $file_systems($fs_id) {*}$node_key Directories]]
}


proc memfs::node_close_handler {fs_id node_key chan} {
    variable file_systems
    node_set_times_now $fs_id $node_key
    dict unset file_systems($fs_id) {*}$node_key Chans $chan
    return
}

package provide memfs 1.0
