# Here is a variation of the simple network server we saw in
# <<sect_network_socket_server>>. The difference here is that while
# that server simply reversed the line sent by the client, this server
# will execute it as an expression, essentially acting as a remote
# calculator. This is obviously dangerous since the `expr`
# command can evaluate arbitrary Tcl scripts. We will therefore have
# to do the evaluation within a safe interpreter.

# We start off by defining the
# command that will invoked in response to every connection request.
# For each new connection, we create a safe interpreter that
# will be dedicated to that connection. Furthermore, we place limits
# on the number of commands that may be executed and the maximum time
# the interpreter is allowed to live. Then
# we make the accepted client channel non-blocking and
# attach a read handler passing it our slave interpreter.

proc on_accept {so client_ip client_port} {
    set slave [interp create -safe]
    set cmdlimit [$slave eval {info cmdcount}]; # Get number of commands executed so far in slave
    $slave limit command -value [incr cmdlimit 3]
    $slave limit time -seconds [expr {[clock seconds] + 60}]
    
    chan configure $so -buffering line -encoding utf-8 -blocking 0 -translation crlf
    chan event $so readable [list on_read $slave $so $client_port]
}
# Notice how we set the command limit. The slave internally
# evaluates commands as part of its internal initialization. So the limit
# we set has to be some increment to the number of commands already executed
# by the slave.

# Our read handler reads each line from the client and evaluates it as
# an expression in the slave. For ease of illustration we do not
# bother dealing with expressions across multiple lines.

proc on_read {slave so client_port} {
    set n [gets $so line]
    if {$n >= 0} {
        set status [catch {$slave eval [list expr $line]} result]
        puts $so $result
        if {$status &&
            [lindex $::errorCode 0] eq "TCL" &&
            [lindex $::errorCode 1] eq "LIMIT"} {
            close $so
            interp delete $slave
        }
    } elseif {[chan eof $so]} {
        close $so
        interp delete $slave
    }
}
# Make a note of the error handling. If the slave evaluation results
# in an error because of limits being exceeded, we close down the connection.
# Other errors like invalid input are simply reported back to the client.

# Now we start up the server on our local address.
set listener [socket -server on_accept -myaddr 127.0.0.1 49999]
vwait forever

# [NOTE]
# There is a design choice we made in our implementation to have the master
# interpreter handle the I/O and invoke the slave only to evaluate the
# expression. An alternative would have been to hand off the channel to
# the slave and have it handle the I/O itself. Our choice was based on
# simpler clean up of the channel when the slave interpreter crossed
# its limits.
